/* global process */

'use strict';

var
    Path = require('path'),
    log = require('gulplog'),
    chalk = require('chalk'),
    yargs = require('yargs'),
    tildify = require('tildify'),
    fancyLog = require('fancy-log'),
    
    options = require('./options'),
    launch = require('./launcher')
;

process.env.INIT_CWD = process.cwd();

var logLevels = [
    'error', // -v: Logs error events.
    'warn', // -vv: Logs warn and error events.
    'info', // -vvv: Logs info, warn and error events.
    'debug' // -vvvv: Logs all log levels.
];

function initLogging(config) {
    if (config.simple || config.silent) {
        log.on('error', () => {});
        return;
    }

    logLevels
        .filter((item, i) => {
            return i < config.verbosity;
        })
        .forEach((level) => {
            if (level === 'error') {
                log.on(level, fancyLog.error);
            } else {
                log.on(level, fancyLog);
            }
        });
}

function getConfig() {
    var usage =
      '\n' + chalk.bold('Usage:') +
      ' huddler ' + chalk.blue('[options]') + ' ...tasks';

    var parser = yargs.usage(usage, options);
    var opts = parser.argv;
    
    var config = {
        taskfile: Path.resolve(opts.taskfile || 'taskfile.js'),
        configFile: Path.resolve(opts.config || 'huddler.json'),
        verbosity: opts.verbosity || 3
    };
    
    if (opts.help) {
        config.help = () => {
            console.log(parser.help());
        };
        return config;
    }
    
    if (opts.list) {
        if (opts.json) {
            config.action = 'json';
            config.json = opts.json;
        } else if (opts.simple) {
            config.action = 'simple';
        } else {
            config.action = 'list';
        }
    } else {
        config.action = 'run';
        config.tasks = opts._;
    }
    
    return config;
}

function run() {
    var config = getConfig();
    var cwd = Path.dirname(config.taskfile);
    
    initLogging(config);
    
    if (config.help) {
        config.help();
        return;
    }
    
    if (process.cwd() !== cwd) {
        process.chdir(cwd);
        log.info(
            'Working directory changed to',
            chalk.magenta(tildify(cwd))
        );
    }
    
    launch(config);
}

module.exports = run;