/* global process */

'use strict';

var
    logger      = require('gulplog'),
    chalk       = require('chalk'),
    prettyTime  = require('pretty-hrtime'),
    stdout      = require('mute-stdout')
;

function exit(code) {
    if (process.platform === 'win32' && process.stdout.bufferSize) {
        process.stdout.once('drain', function () {
            process.exit(code);
        });
        return;
    }
    process.exit(code);
}

function formatError(evt) {
    if (!evt.error) {
        return evt.message;
    }

    // PluginError
    if (typeof evt.error.showStack === 'boolean') {
        return evt.error.toString();
    }

    // Normal error
    if (evt.error.stack) {
        return evt.error.stack;
    }

    // Unknown (string, number, etc.)
    return new Error(String(evt.error)).stack;
}

function loadTasks(config) {
    logger.debug('Loading taskfile');
    
    var exported = require(config.taskfile);
    var taskNames = Object.keys(exported);
    
    logger.debug('Loading huddler module');
    var huddler = require('../..');

    if (taskNames.length) {
        taskNames.forEach((taskName) => {
            var task = exported[taskName];

            if (typeof task !== 'function') {
                return;
            }

            huddler.task(taskName, task);
        });
    }
    
    return huddler;
}

function registerEventListeners(huddler) {
    var loggedErrors = [];
    var runningTasks = {};

    huddler.on('start', (evt) => {
        // TODO: batch these
        // so when 5 tasks start at once it only logs one time with all 5
        var level = evt.branch ? 'debug' : 'info';
        
        runningTasks[evt.uid] = evt.name;
        
        logger[level](
            'Starting', '\'' + chalk.cyan(evt.name) + '\'...'
        );
    });

    huddler.on('stop', (evt) => {
        var time = prettyTime(evt.duration);
        var level = evt.branch ? 'debug' : 'info';
        
        delete runningTasks[evt.uid];
        
        logger[level](
            'Finished', '\'' + chalk.cyan(evt.name) + '\'',
            'after', chalk.magenta(time)
        );
    });

    huddler.on('error', (evt) => {
        var msg = formatError(evt);
        var time = prettyTime(evt.duration);
        var level = evt.branch ? 'debug' : 'error';
        
        delete runningTasks[evt.uid];
        
        logger[level](
            '\'' + chalk.cyan(evt.name) + '\'',
            chalk.red('errored after'),
            chalk.magenta(time)
        );

        // If we haven't logged this before, log it and add to list
        if (loggedErrors.indexOf(evt.error) === -1) {
            logger.error(msg);
            loggedErrors.push(evt.error);
        }
    });
    
    process.once('exit', () => {
        var taskKeys = Object.keys(runningTasks);
        if (!taskKeys.length) {
            return;
        }

        var taskNames = taskKeys.map(function (key) {
            return runningTasks[key];
        }).join(', ');

        logger.warn(
            chalk.red('The following tasks did not complete:'),
            chalk.cyan(taskNames)
        );

        logger.warn(
            chalk.red('Did you forget to signal async completion?')
        );
    });
}

function onError(err) {
    if (err) {
    throw err;
        logger.error(chalk.red(err.message));
        logger.error('To specify a taskfile, try running: huddler --taskfile <path>');
        exit(1);
    }
}

function execute(config) {
    if ('run' !== config.action) {
        // Mute stdout if we are listing tasks
//        stdout.mute();
    }
    
    try {
        var huddler = loadTasks(config);
    } catch(err) {
        return onError(err);
    }
    
    registerEventListeners(huddler);
    
    stdout.unmute();
    
    process.nextTick(() => {
        var handler = require('../tasks/' + config.action);
        
        handler(huddler, config, onError);
    });
}

module.exports = execute;
