'use strict';

var
    Path        = require('path'),
    fs          = require('fs'),
    
    replace     = require('lodash.replace'),
    kebabCase   = require('lodash.kebabcase'),
    isFunction  = require('lodash.isfunction'),
    forEach     = require('lodash.foreach'),
    splitWords  = require('lodash.words'),
    logger      = require('gulplog'),
    
    Bundle      = require('./bundle'),
    builtIn     = require('../tasks/default'),
    cache       = require('./cache')
;

function isNumeric(num) {
    return !isNaN(num);
}

function resolve(name, paths) {
    for (var i = 0; i < paths.length; i++) {
        if (fs.existsSync(paths[i])) {
            return paths[i];
        }
    }
    
    throw new Error('Failed to resolve bundle "' + name + '"');
}

function friendlyName(bundle) {
    bundle = replace(bundle, /Bundle$/, '');
    return kebabCase(bundle);
}

var loaded = false;

function loadBundles(huddler)
{
    if (loaded) {
        return;
    }
    
    loaded = true;
    
    var oldTask = huddler.task;
    var bundles = huddler.getOption('bundles');
    
    var src_dir = Path.resolve(huddler.getOption('src_dir'));
    var task_path = huddler.getOption('task_path');
    
    forEach(bundles, (bundle_path, name) => {
        if (isNumeric(name)) {
            name = bundle_path;
            bundle_path = Path.join(src_dir, name);
            
            if (!fs.existsSync(bundle_path)) {
                var words = splitWords(replace(bundle_path, /Bundle$/, ''), /[A-Z][^A-Z]+/g);
                var possible = [];
                
                if (2 === words.length) {
                    possible.push(Path.join(src_dir, words[0], words[1] + 'Bundle'));
                    possible.push(Path.join(src_dir, words[0], 'Bundle', + words[1] + 'Bundle'));
                } else if (1 === words.length) {
                    possible.push(Path.join(src_dir, words[0] + 'Bundle'));
                }
                
                bundle_path = resolve(name, possible);
            }
        }
        
        logger.debug('Preparing to load bundle "' + name + '"');
        
        let bundle_name = friendlyName(name);
        
        huddler.task = function (name, callback) {
            if (!cache[name]) {
                cache[name] = new Map();
            }

            let bundle = new Bundle(name, callback);
            let task_name = bundle_name + ':' + name;

            // Add the bundle task to undertaker under a prefixed name
            bundle.register(oldTask, task_name);

            cache[name].set(task_name, bundle);
        };

        let task_file = Path.join(bundle_path, task_path);
        let exported = builtIn;

        try {
            exported = require(task_file);
        } catch (err) {
            if (err.code !== 'MODULE_NOT_FOUND' || -1 === err.message.indexOf(task_path)) {
                throw err;
            }
            logger.debug('Taskfile "' + task_file + '" not found. Using default.');
        }

        if (isFunction(exported)) {
            exported(huddler, bundle_name, bundle_path);
        } else {
            forEach(exported, (task, name) => {
                if (!isFunction(task)) {
                    throw new Error('A taskfile may export a single function or hash of functions');
                }
                huddler.task(name, task);
            });
        }
    });
    
    logger.debug('Restoring old task function');
    
    huddler.task = oldTask;
};

module.exports = loadBundles;