'use strict';

var
    json = require('jsonfile'),
    definitions = require('./options')
;

function is (type, value)
{
    switch (type) {
        case 'null':
            return null === value;
        case 'array':
            return Array.isArray(value);
        default:
            return typeof value === type;
    }
};

function valueOf (value)
{
    if (value != null && typeof value === 'object' && typeof value.valueOf === 'function') {
        value = value.valueOf();
    }
    
    return value;
};

function validate (type, value) {
    return is(type, value) ? value : null;
};

var operators = {
    string: (value) => {
        if (value != null && typeof value === 'object' && typeof value.toString === 'function') {
            value = value.toString();
        }
        return is('string', value) ? value : null;
    },
    number: (value) => {
        return validate('number', valueOf(value));
    },
    boolean: (value) => {
        return validate('boolean', valueOf(value));
    },
    date: (value) => {
        value = valueOf(value);
        if (typeof value === 'number' && !isNaN(value) && isFinite(value)) {
            return new Date(value);
        }
        return null;
    },
    object: (value) => {
        return validate('object', value);
    },
    array: (value) => {
        return validate('array', value);
    }
};

function normalize (type, value)
{
    if (is('string', type)) {
        if (operators[type]) {
            return operators[type](value);
        }
        throw new Error('Unknown operator "' + type + '"');
    }

    if (is('function', type)) {
        return type(value);
    }
    
    if (is('array', type)) {
        for (var i = 0; i < type.length; i++) {
            let result = normalize(type[i], value);
            if (result) {
                return result;
            }
        }
        return null;
    }
    
    throw new Error('Expected a string, array or function');
};

function Config(cached)
{
    this._definitions = definitions;
    this._cache = cached;
    this._resolver = null;
};

Config.prototype._initialize = function(options)
{
    if (!this._cache) {
        this._cache = {};
    } else if (is('string', this._cache)) {
        // This throws an exception by default
        this._cache = json.readFileSync(this._cache);
    }
    
    return Object.assign({}, this._cache, options);
};

Config.prototype.get = function(key)
{
    if (!this._resolver) {
        this._resolver = this.resolve({});
    }
    
    return this._resolver.get(key);
};

Config.prototype.resolve = function(options)
{
    var call_stack = [];
    
    options = this._initialize(options);
    
    var definitions = this._definitions;
    var resolver = {
        get: fetch
    };
    
    function fetch (key) {
        var option = definitions[key];
        if (!option) {
            throw new Error('option "' + key + '" doesn\'t exist');
        }

        if (-1 !== call_stack.indexOf(key)) {
            throw new Error('Recursive option retrieval not permitted.');
        }
        
        call_stack.push(key);
        
        function get (value, args) {
            if (is('function', value)) {
                value = value.apply(resolver, args);
            }

            return normalize(option.type, value);
        };

        var args = Array.prototype.slice.call(arguments, 1);
        var result = get(options[key], args);
        
        if (!result) {
            result = get(option.default, args);
        }
        
        call_stack.pop();

        return result;
    };
    
    return resolver;
};

module.exports = Config;