'use strict';

var
    Path = require('path'),
            
    through = require('through2'),
    pumpify = require('pumpify'),
    lead = require('lead'),
    sourcemap = require('vinyl-sourcemap'),
    
    currentTask = require('../tasks').current,
    filesystem = require('../util/filesystem'),
    hash = require('../util/hasher'),
    Tracker = require('../util/tracker')
;
    
function prepareFiles(options) {
    // Todo do options need to be passed to through2
    return through.obj(options, function(file, enc, cb) {
        // Exclude all directories from passing through
        if (!file.isDirectory()) {
            file.stat = (file.stat || new filesystem.Stats());
            file.stat.mode = options.get('mode', file);
            file.flag = options.get('flag', file);
            file.cwd = options.web_dir;
            file.path = Path.join(options.dest, file.basename);
            file.base = options.dest;

            filesystem.makeDirectory(options.dest, options.get('web_mode'), (err) => {
                if (err) {
                    return cb(err);
                }
                cb(null, file);
            });
        } else {
            cb();
        }
    });
};

function createRenamer (options, file) {
    var rename = options.get('rename');
    
    if (rename && typeof rename !== 'function') {
        rename = (function (){
            var bandage = 1;
            var stem = file.stem;

            return function(file) {
                file.stem = stem + '_' + bandage++;
            };
        })();
    }
    
    return rename;
};
    
function symlinkFiles(options) {
    return through.obj(function (file, enc, cb) {
        // Symlinks can only be created for files which have not been
        // explicitly converted to a symlink or contain no data.
        if (!options.get('symlink', file) || file.symlink || !file.hash) {
            return cb(null, file);
        }
        
        if (file.hash !== hash(file)) {
            return cb(null, file);
        }

        var target = file.history[0];
        var rename = createRenamer(options, file);
        
        if (options.get('relative')) {
            target = Path.relative(file.base, target);
        }

        filesystem.makeSymlink(file, target, handler);
        
        function handler(err) {
            if (rename && err && err.code === 'EEXIST') {
                rename(file);
                return filesystem.makeSymlink(file, target, handler);
            }

            cb(err);
        };
    });
};

function sourcemapFiles(options) {
    return through.obj(function(file, enc, cb) {
        var target = options.get('sourcemaps', file);
        if (!target) {
            return cb(null, file);
        }
        
        if (typeof target !== 'string') {
            target = undefined;
        }

        sourcemap.write(file, target, (err, updatedFile, sourcemapFile) => {
            if (err) {
                return cb(err);
            }

            this.push(updatedFile);
            if (sourcemapFile) {
                this.push(sourcemapFile);
            }

            cb();
        });
    });
};
    
function writeFiles(options) {
    return through.obj(function(file, enc, cb) {
        var rename = createRenamer(options, file);

        filesystem.writeFile(file, handler);

        function handler(err) {
            if (rename && err && err.code === 'EEXIST') {
                rename(file);
                return filesystem.writeFile(file, handler);
            }

            cb(err);
        };
    });
};

function webDir(options)
{
    var cwd = options.get('cwd');
    var web_dir = Path.resolve(cwd, options.get('web_dir'));
    var targets = options.get('targets');
    var tracker = new Tracker(web_dir, Path.resolve(cwd, options.get('tracker')));
    
    return function (opt) {
        if (typeof opt === 'string') {
            opt = {sub_dir: opt};
        }
        
        opt = options.resolve(opt);
        
        var sub_dir = opt.get('sub_dir') || targets[currentTask()] || '';
        var curr_task = currentTask();
        
        opt.dest = Path.resolve(web_dir, sub_dir);
        opt.web_dir = web_dir;
        
        if (-1 === opt.dest.indexOf(web_dir)) {
            throw new Error('Cannot write outside web dir "' + opt.dest + '"');
        }

        var saveStream = pumpify.obj(
            prepareFiles(opt),
            tracker.track(curr_task),
            symlinkFiles(opt),
            sourcemapFiles(opt),
            writeFiles(opt)
        );

        saveStream.once('end', () => {
            tracker.dump((err) => {
                if (err) {
                    saveStream.emit('error', err);
                }
            });
        });

        // Sink the output stream to start flowing
        return lead(saveStream);
    };
};

module.exports = webDir;