'use strict';

var
    fs = require('fs'),
    dirname = require('path').dirname
;

function isFatalError(err, file) {
    if (!err) {
        return false;
    }

    if (err.code === 'EEXIST' && (!file || file.flag === 'wx')) {
        // Handle scenario for file overwrite failures.
        return false;   // "These aren't the droids you're looking for"
    }

    // Otherwise, this is a fatal error
    return true;
};

function isInt(value) {
    var x;
    if (isNaN(value)) {
        return false;
    }
    x = parseFloat(value);
    return (x | 0) === x;
}

function writeStream(path, file, cb) {
    var opt = {
        mode: file.stat.mode,
        flags: file.flag
    };

    var outStream = fs.createWriteStream(path, opt);

    file.contents.once('error', done);
    file.contents.once('end', done);
    outStream.once('error', done);
    outStream.once('finish', done);

    file.contents.pipe(outStream);

    function done(err) {
        file.contents.removeListener('error', done);
        file.contents.removeListener('end', done);
        outStream.removeListener('error', done);
        outStream.removeListener('finish', done);

        outStream.end((endErr) => {
            cb(err || endErr);
        });
    }
};

function writeBuffer(path, file, cb) {
    var opt = {
        mode: file.stat.mode,
        flag: file.flag
    };

    fs.writeFile(path, file.contents, opt, cb);
}

function makeDirectory(path, mode, cb) {
    if (typeof mode === 'function') {
        cb = mode;
        mode = null;
    } else if (typeof mode === 'string') {
        mode = parseInt(mode, 8);
    } else if (!isInt(mode)) {
        mode = parseInt('0777', 8) & (~process.umask());
    }
    
    fs.mkdir(path, mode, (err) => {
        if (err) {
            if (err.code === 'ENOENT') {
                makeDirectory(dirname(path), mode, (err) => {
                    if (!err || err.code === 'EEXIST') {
                        return makeDirectory(path, mode, cb);
                    }
                    cb(err);
                });
            } else {
                fs.stat(path, function (statErr, stat) {
                    if (statErr || !stat.isDirectory()) {
                        return cb(err);
                    }
                    cb();
                });
            }
        } else {
            cb();
        }
    });
};

function symlink(file, target, cb) {
    if (typeof target === 'function') {
        cb = target;
        target = file.symlink;
    }
    
    var source = file.path || file;
    var type = file.isDirectory && file.isDirectory() ? 'dir' : 'file';
    
    if (file.flag === 'w') {
        fs.unlink(source, (err) => {
            if (!err|| err.code === 'ENOENT') {
                return fs.symlink(target, source, type, cb);
            }
            cb(err);
        });
    } else {
        fs.symlink(target, source, type, cb);
    }
};

function writeFile(file, cb) {
    if (file.isDirectory()) {
        return makeDirectory(file.path, file.stat.mode, cb);
    }

    // Stream it to disk yo
    if (file.isStream()) {
        return writeStream(file.path, file, cb);
    }

    // Write it as a symlink
    // file.isSymbolic() will only check original file
    // Only a plugin will configure the file.symlink. gulp.src will
    // resolve them to their full paths.
    if (file.symlink) {
        return symlink(file, file.symlink, cb);
    }

    // Write it like normal
    if (file.isBuffer()) {
        return writeBuffer(file.path, file, cb);
    }

    // If no contents then do nothing
    if (file.isNull()) {
        return cb();
    }
};

module.exports = {
    writeFile: writeFile,
    makeDirectory: makeDirectory,
    makeSymlink: symlink,
    Stats: fs.Stats
};